
from ufl.classes import Terminal, UtilityType
from ufl.algorithms import Graph, expand_indices
from uflacs.codeutils.format_code_structure import format_code_structure
from uflacs.codeutils.expr_formatter import ExprFormatter
from uflacs.codeutils.cpp_format import CppFormatterRules
from uflacs.codeutils.target_formatter import CppDefaultFormatter

def compile_element(element, prefix):
    return "// TODO: Compile element %s" % str(element)

def compile_expression(expr, prefix):
    return "// TODO: Compile expression %s" % str(expr)

def compile_form(form, prefix):

    # This formatter is a multifunction implementing target
    # specific formatting rules, here using the default rules.
    target_formatter = CppDefaultFormatter()

    # This formatter is a multifunction with single operator
    # formatting rules for generic C++ formatting
    cpp_formatter = CppFormatterRules(target_formatter)

    # First we preprocess the form in standard UFL fashion
    fd = form.compute_form_data()

    # We'll place all code in a list while building the program
    code = []

    # Then we iterate over the integrals
    for data in fd.integral_data:
        domain_type, domain_id, integrals, metadata = data
        for itg in integrals:
            # Fetch the expression
            integrand = itg.integrand()

            # Then we apply the additional expand_indices preprocessing that form preprocessing does not
            expr = expand_indices(integrand)

            # And build the computational graph of the expression
            G = Graph(expr)
            V, E = G

            # In this dictionary we will place ufl expression to C++
            # variable name mappings while building the program
            variables = {}

            # This final formatter implements a generic framework handling indices etc etc.
            expr_formatter = ExprFormatter(cpp_formatter, variables)
            integral_code = [] # TODO: Use code formatting utils

            nv = len(V)
            vnum = 0
            for i, v in enumerate(V):
                # Check if we should make a variable here
                if i == nv-1:
                    vname = 'integrand'
                elif (not isinstance(v, UtilityType)
                      and v.shape() == ()
                      and not isinstance(v, Terminal)):
                    vname = 'v%d' % vnum
                    vnum += 1
                else:
                    vname = None
                # If so, generate code for it
                if vname is not None:
                    vcode = expr_formatter.visit(v)
                    integral_code.append("%s = %s;" % (vname, vcode))
                    expr_formatter.variables[v] = vname

            # Join code to what we have
            code.append(integral_code)

    return format_code_structure(code)
