
import os
from uflacs.utils.log import info
from uflacs.utils.files import write_file

def add_compile_cpp1_options(opts):
    "Args: list of .ufl file(s)."
    pass

def run_compile_cpp1(options, args):
    "Compile forms and expressions from .ufl file(s) into C++ code with undefined FEM backend."
    from ufl.algorithms import load_ufl_file
    from uflacs.backends.cpp1.compiler import compile_element, compile_expression, compile_form
    from uflacs.codeutils.format_code_structure import format_code_structure

    for input_filename in args:
        prefix, ext = os.path.splitext(os.path.basename(input_filename))
        if ext != '.ufl':
            warning("Expecting ufl file, got %s." % ext)
        output_filename = prefix + '.h'

        info("Loading file '%s'..." % (input_filename,))
        data = load_ufl_file(input_filename)

        info("Compiling '%s'..." % prefix)
        code = format_code_structure([
            ['// Elements'],
            [compile_element(element, prefix) for element in data.elements],
            ['// Expressions'],
            [compile_expression(expr, prefix) for expr in data.expressions],
            ['// Forms'],
            [compile_form(form, prefix) for form in data.forms],
            ])

        info("Writing code to '%s'..." % output_filename)
        write_file(output_filename, code)

    return 0

